// -*- C++ -*-
//
//  This file is part of Siena, a wide-area event notification system.
//  See http://www.cs.colorado.edu/serl/siena/
//
//  Authors: Antonio Carzaniga <carzanig@cs.colorado.edu>
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 2002 University of Colorado
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
//
// $Id: bitvector.h,v 1.5 2002/09/30 04:51:48 carzanig Exp $
//
#ifndef BITVECTOR_H
#define BITVECTOR_H

#include <cstddef> /* for size_t */

#include "fwdconf.h"
#include "ft_allocator.h"

class bitvector;
class ibitvector;

typedef unsigned long		bv_atom_t;
//
// hopefully this stuff is statically optimized by the compiler
//
static const unsigned char	bv_atom_bits = sizeof(bv_atom_t) * 8;
static const unsigned char	bv_atom_bits_mask = bv_atom_bits - 1;
static const unsigned char	bv_atom_bits_shift 
= (bv_atom_bits == 32 ? 5 : 
   (bv_atom_bits == 64 ? 6 :
    (bv_atom_bits == 16 ? 4 : 
     /* bv_atom_bits == 16 */ 2)));

/** fixed-size bit vector.
 *
 *  This bitvector is implemented by a single array.
 **/
class bitvector {
public:
    /** constructs a bitvector capable of holding <em>size</em> bits
     **/
				bitvector(size_t size);
				~bitvector();
    
    bool			test(size_t pos)			const;
    bool			set(size_t pos);

    void			set(const ibitvector & x);

    void			clear();
    /** number of bits set to 1 **/
    size_t			get_count()				const;
    /** total bit capacity **/
    size_t			get_size()				const;

private:
    bv_atom_t *			elements; 
    const size_t		size;		// number of bits
    size_t			count;		// number of bits set to 1

    size_t			element_size()				const;

    static size_t		set(bv_atom_t * x, bv_atom_t * xe, 
				    const bv_atom_t * y, const bv_atom_t *ye);
};

/** variable-size bit vector.
 *
 *  This bit vector is implemented by a multi-index structure.
 **/
class ibitvector {
public:
				ibitvector();
    
    bool			test(size_t pos)			const;
    bool			set(size_t pos, FTAllocator &);

    void			clear();
    size_t			get_count()				const;
    size_t			get_size()				const;

private:
    struct index;
    struct block;

    static const unsigned int	block_size = 16;
    static const unsigned int	block_shift = 4;
    static const unsigned int	block_mask = block_size - 1;

    static const unsigned int	index_size = 16;
    static const unsigned int	index_shift = 4;
    static const unsigned int	index_mask = index_size - 1;

    union index_or_block {
	block *		b;
	index *		i;
    };

    struct block {
	index *		up;
	bv_atom_t	elements[block_size];
	
	block(index *);
    };

    struct index {
	index *		up;
	index_or_block	down[index_size];

	index(index *);
    };

    class iterator {
    public:
			iterator(const block &);

	size_t		element_address() const;
	const block *	next_block();

    private:
	index *		bi;
	size_t		addr;
	unsigned char	level;
    };

    size_t count;
    size_t size;
    block  first_block;

    friend class bitvector;
};

#ifdef HAVE_INLINE
#include "bitvector.icc"
#endif

#endif
