// -*- C++ -*-
//
//  This file is part of Siena, a wide-area event notification system.
//  See http://www.cs.colorado.edu/serl/siena/
//
//  Authors: Antonio Carzaniga <carzanig@cs.colorado.edu>
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 2002 University of Colorado
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
//
// $Id: siena_types.cc,v 1.4 2002/09/30 04:50:25 carzanig Exp $
//
#include <iostream>
#include "fwdconf.h"
#include "siena_types.h"

bool covers(const sx_predicate & p, const sx_message & m) {
    for(sx_predicate::const_iterator i = p.begin(); i != p.end(); ++i)
	if (covers(*i, m)) return true;
    return false;
}

bool covers(const sx_filter & f, const sx_message & m) {
    sx_filter::const_iterator i = f.begin();
    if (i == f.end()) return false;
    do {
	if (!covers(*i, m)) return false;
    } while (++i != f.end());
    return true;
}

bool covers(const sx_constraint & c, const sx_message & m) {
    sx_message::const_iterator i = m.find(c.name);
    if (i == m.end()) return false;
    if ((*i).second.type() != c.value.type()) return false;
    switch ((*i).second.type()) {
    case SX_string: {
	const string & x = (*i).second.string_value();
	const string & v = c.value.string_value();
	switch (c.op) {
	case SX_eq: return x == v;
	case SX_gt: return x > v;
	case SX_lt: return x < v;
	case SX_pf: return x.find(v,0) == 0;
	case SX_sf: return x.size() >= v.size() 
			&& x.rfind(v, 0) == 0;
	case SX_ss: return x.size() >= v.size() 
			&& x.find(v, x.size()-v.size()) < x.size();
	}
    }
    case SX_int: 
	switch (c.op) {
	case SX_eq: return (*i).second.int_value() == c.value.int_value();
	case SX_gt: return (*i).second.int_value() > c.value.int_value();
	case SX_lt: return (*i).second.int_value() < c.value.int_value();
	}
    case SX_double: 
	switch (c.op) {
	case SX_eq: return (*i).second.double_value() == c.value.double_value();
	case SX_gt: return (*i).second.double_value() > c.value.double_value();
	case SX_lt: return (*i).second.double_value() < c.value.double_value();
	}
    case SX_bool: 
	switch (c.op) {
	case SX_eq: return (*i).second.bool_value() == c.value.bool_value();
	case SX_gt: return (*i).second.bool_value() > c.value.bool_value();
	case SX_lt: return (*i).second.bool_value() < c.value.bool_value();
	}
    }
}

ostream & operator << (ostream & os, const sx_value & x) {
    switch (x.type()) {
    case SX_int: return os << x.int_value();
    case SX_string: return os << '"' << x.string_value() << '"';
    case SX_bool: return os << x.bool_value();
    case SX_double: return os << x.double_value();
    case SX_null: return os << "null";
    default:
	return os;
    }
}

static const char * sx_op_table[] = {
    "", "=", "<", ">", ">=", "<=", "*<", "*>", "@", "!=", "**"
};

ostream & operator << (ostream & os, const sx_constraint & x) {
    return os << x.name << ' ' << sx_op_table[x.op] << ' ' << x.value;
}

ostream & operator << (ostream & os, const sx_message & m) {
    sx_message::const_iterator mi = m.begin();
    if (mi != m.end()) {
	os << (*mi).first << " = " << (*mi).second;
	while(++mi != m.end())
	    os << ' ' << (*mi).first << " = " << (*mi).second;
    }
    return os;
}

ostream & operator << (ostream & os, const sx_filter & f) {
    sx_filter::const_iterator fi = f.begin();
    if (fi != f.end()) {
	os << *fi;
	while(++fi != f.end())
	    os << ", " << *fi;
    }
    return os;
}

ostream & operator << (ostream & os, const sx_predicate & p) {
    sx_predicate::const_iterator pi = p.begin();
    if (pi != p.end()) {
	os << *pi;
	while(++pi != p.end())
	    os << endl << " | " << *pi;
    }
    return os;
}
