// -*- C++ -*-
//
//  This file is part of Siena, a wide-area event notification system.
//  See http://www.cs.colorado.edu/serl/siena/
//
//  Authors: Antonio Carzaniga <carzanig@cs.colorado.edu>
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 2001-2002 University of Colorado
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
//
// $Id: tst.cc,v 1.15 2002/09/29 15:02:48 carzanig Exp $
//
#include <iostream>

#include "fwdconf.h"
#include "tst.h"
#include "ft_allocator.h"

#ifndef HAVE_INLINE
#include "tst.icc"
#endif

void TSTNode::print(int level) const {
    int i;
    if (left != NULL) {
	cout << "    "; left->print(level+1); 
	for(i = 0; i<level; ++i) cout << "    "; cout << "  /" << endl;
	for(i = 0; i<level; ++i) cout << "    ";
    }
    cout << " -";
    if ( c == '\0' ) {
	cout << this;
    } else {
	cout << c;
    }
    cout << (data == NULL ? ' ' : '+' );
    if (middle != NULL) {
	middle->print(level+1);
    } else {
	cout << endl;
    }
    if (right != NULL) {
	for(i = 0; i<level; ++i) cout << "    "; cout << "  \\" << endl;
	for(i = 0; i<level; ++i) cout << "    "; cout << "    ";
	right->print(level+1);
    }
}

void TST::print() const {
    if (root != NULL) root->print();
}

void TST::clear() {
    //
    // I don't actually deallocate stuff here since I'm assuming that
    // ftmemory will be deallocated all at once
    //
    root = NULL;
}

//
// these two algorithms are taken more or less directly from
// R. Sedgewick's "Algorithms in C" 3rd Ed. pp 638--639.
//
TSTNode * TST::insert(const char * s, FTAllocator & ftmemory) {
    TSTNode * pp = NULL;
    TSTNode ** p = &root;
    while (*p != NULL) {
	pp = *p;
        if (*s == pp->c) {
            if (*s == '\0') return pp;
	    ++s;
            p = &(pp->middle);
        } else if (*s < pp->c) {
            p = &(pp->left);
	} else {
            p = &(pp->right);
	}
    }

    for (;;) {
	*p = new (ftmemory)TSTNode(*s, pp);
	pp = *p;
        if (*s == '\0') return pp;
	++s;
        p = &(pp->middle);
    }
}

const TSTNode * TST::find(const char * s) const {
    const TSTNode * p = root;
    while (p != NULL) {
        if (*s < p->c) {
            p = p->left;
	} else if (*s == p->c)  {
            if (*s == '\0')
                return p;
	    ++s;
            p = p->middle;
        } else {
            p = p->right;
	}
    }
    return NULL;
}

//
// these algorithms are not in R. Sedgewick's "Algorithms in C"
//
TSTNode * TSTNode::backtrack_next() const {
    TSTNode * pp;
    const TSTNode * p = this;
    while ((pp = p->up) != NULL) {
	if (p == pp->left) goto back_from_left;
	if (p == pp->middle) goto back_from_middle;
	if (p == pp->right) goto back_from_right;
	return NULL;	// this should NEVER happen, but...
    back_from_left:
	if (pp->c == '\0') return pp; // probably useless
	if (pp->middle) return pp->middle;
    back_from_middle:
	if (pp->right) return pp->right;
    back_from_right:
	p = pp;
    }
    return pp;
}

TSTNode * TSTNode::next() const {
    TSTNode * p;
    if (left) {
	p = left;
    } else if (middle) {
	p = middle;
    } else if (right) {
	p = right;
    } else {
	if ((p = backtrack_next()) == NULL)
	    return NULL;
    }
    for (;;) {
	if (p->left) {
	    p = p->left;
        } else if (p->c == '\0') {
	    return p; 
	} else if (p->middle) {
            p = p->middle;
	} else if (p->right) {
	    p = p->right;
	} else {
	    if ((p = p->backtrack_next()) == NULL)
		return NULL;
	}
    }
}

TSTNode * TSTNode::backtrack_prev() const {
    TSTNode * pp;
    const TSTNode * p = this;
    while ((pp = p->up) != NULL) {
	if (p == pp->left) goto back_from_left;
	if (p == pp->middle) goto back_from_middle;
	if (p == pp->right) goto back_from_right;
	// this should NEVER happen, but just to be safe: 
	return NULL;
    back_from_right:
	if (pp->c == '\0') return pp;
	if (pp->middle != NULL) return pp->middle;
    back_from_middle:
	if (pp->left != NULL) return pp->left;
    back_from_left:
	p = pp;
    }
    return pp;
}

TSTNode * TSTNode::prev() const {
    TSTNode * p;
    if (left != NULL) {
	p = left;
    } else {
	if ((p = backtrack_prev()) == NULL)
	    return NULL;
	if (p->c == '\0') return p;
    }
    for(;;) {
	if (p->right) {
	    p = p->right;
	} else if (p->middle) {
	    p = p->middle;
	} else if (p->c == '\0') {
	    return p; 
	} else if (p->left) {
	    p = p->left;
	} else {
	    if ((p = p->backtrack_prev()) == NULL)
		return NULL;
	    if (p->c == '\0') return p;
	}
    }
}

TSTNode * TST::first() const {
    TSTNode * pp = root;
    if (pp == NULL) return NULL;
    for(;;) {
	if (pp->left) {
	    pp = pp->left;
        } else if (pp->c == '\0') {
	    return pp; 
	} else if (pp->middle) {
            pp = pp->middle;
	} else if (pp->right) {
	    pp = pp->right;
	} else {
	    //
	    // this should never happen
	    //
	    return NULL;
	}
    }
}

TSTNode * TSTNode::upper_bound(char s) const {
    if (s <= c) return next();
    TSTNode * p;
    if ((p = backtrack_next()) == NULL)
	return NULL;
    for (;;) {
	if (p->left) {
	    p = p->left;
        } else if (p->c == '\0') {
	    return p; 
	} else if (p->middle) {
            p = p->middle;
	} else if (p->right) {
	    p = p->right;
	} else {
	    if ((p = p->backtrack_next()) == NULL)
		return NULL;
	}
    }
}

TSTNode * TSTNode::lower_bound(char s) const {
    TSTNode * p;
    //
    // pre-condition: c != *s
    //
    if (c < s) {
	if (middle != NULL) {
	    p = middle;
	    if (p->c == '\0') return p;
	} else if (left != NULL) {
	    p = left;
	} else {
	    if ((p = backtrack_prev()) == NULL)
		return NULL;
	    if (p->c == '\0') return p;
	}
    } else {
	if ((p = backtrack_prev()) == NULL)
	    return NULL;
	if (p->c == '\0') return p;
    }
    for(;;) {
	if (p->right) {
	    p = p->right;
	}  else if (p->middle) {
	    p = p->middle;
	}else if (p->c == '\0') {
	    return p; 
	} else if (p->left) {
	    p = p->left;
	} else {
	    if ((p = p->backtrack_prev()) == NULL)
		return NULL;
	    if (p->c == '\0') return p;
	}
    }
}

PartialResult PTST::find_partial(const TSTNode ** pp, const char ** s) const {
    const TSTNode * p = *pp;
    if (p == NULL) {
	p = root;
    } else {
	p = p->middle;
    }
    while (p != NULL) {
	*pp = p;
        if (**s < p->c) {
            p = p->left;
	} else if (**s == p->c)  {
            if (**s == '\0') return Found_Complete;
	    ++(*s);
	    if (p->data != NULL) return Found_Partial;
            p = p->middle;
        } else {
            p = p->right;
	}
    }
    return Not_Found;
}

TSTNode * PTST::insert_partial(const char * s, FTAllocator & ftmemory) {
    TSTNode * pp = NULL;
    TSTNode ** p = &root;
    while (*p != NULL) {
	pp = *p;
        if (*s == pp->c) {
	    ++s;
            if (*s == '\0') return pp;
            p = &(pp->middle);
        } else if (*s < pp->c) {
            p = &(pp->left);
	} else {
            p = &(pp->right);
	}
    }
    for (;;) {
	*p = new (ftmemory)TSTNode(*s, pp);
	pp = *p;
	++s;
        if (*s == '\0') return pp;
        p = &(pp->middle);
    }
}
